
from ... conditional import ordereddict


class namespace_cls:
    """
    similar to types.SimpleNamespace
    """
    def __init__(self, kwargs):
        if not isinstance(kwargs, ordereddict):
            kwargs = ordereddict(kwargs)

        self.__dict__ = kwargs

    def __iter__(self):
        return (v for v in self.__dict__.values())

    def __eq__(self, other):
        if (not hasattr(self, '__dict__')) or (not hasattr(other, '__dict__')):
            return NotImplemented

        return self.__dict__ == other.__dict__

    def __getitem__(self, name):
        if name in self.__dict__:
            return self.__dict__[name]

        raise AttributeError("attribute: '{}' not found".format(name))

    def __setitem__(self, name, value):
        self.__dict__[name] = value

    def __len__(self):
        return len(self.__dict__)

    def __repr__(self):
        items = ['{}={!r}'.format(k, v) for k, v in self.__dict__.items()]
        items = ', '.join(items)
        items = items.replace("'", '"')
        items = '{' + items + '}'

        return items





import csv
import gc
import os
import pickle
import shutil

from collections import namedtuple
from datetime import date
from datetime import datetime
from functools import lru_cache
from glob import glob
from io import StringIO
from os.path import isdir  as os_isdir
from os.path import isfile as os_isfile
from urllib.parse import urlparse
from urllib.request import urlopen

from ..conditional import is_windows_os
from ..conditional import ultrajson_installed

if ultrajson_installed:
    # noinspection PyUnresolvedReferences
    import ujson as json
else:
    import json

pickle_extensions = {'.flux',
                     '.pkl',
                     '.pickle'}


def read_file(path,
              encoding=None,
              mode='r',
              filetype=None,
              **kwargs):

    (path,
     encoding,
     filetype,
     mode,
     kwargs) = __validate_io_arguments(path,
                                       encoding,
                                       mode,
                                       filetype,
                                       kwargs,
                                       'read')
    is_url = is_path_a_url(path)

    gc_enabled   = gc.isenabled()
    if gc_enabled: gc.disable()

    if filetype == '.csv':
        data = __read_csv(path, mode, encoding, kwargs)

    elif filetype == '.json':
        data = __read_json(path, mode, encoding, kwargs)

    elif is_url:
        data = __url_request(path, encoding=encoding)

    elif filetype in pickle_extensions:
        with open(path, mode) as f:
            data = pickle.load(f, **kwargs)

    else:
        with open(path, mode, encoding=encoding) as f:
            data = f.read()

    if gc_enabled: gc.enable()

    return data


def write_file(path,
               data,
               encoding=None,
               mode='w',
               filetype=None,
               **kwargs):

    (path,
     encoding,
     filetype,
     mode,
     kwargs) = __validate_io_arguments(path,
                                       encoding,
                                       mode,
                                       filetype,
                                       kwargs,
                                       'write',
                                       is_data_bytes=isinstance(data, bytes))

    gc_enabled   = gc.isenabled()
    if gc_enabled: gc.disable()

    if filetype == '.csv':
        newline = kwargs.pop('newline')
        with open(path, mode, encoding=encoding, newline=newline) as f:
            csv.writer(f, **kwargs).writerows(data)

    elif filetype == '.json':
        with open(path, mode, encoding=encoding) as f:
            json.dump(data, f, **kwargs)

    elif filetype in pickle_extensions:
        with open(path, mode) as f:
            pickle.dump(data, f, **kwargs)

    else:
        with open(path, mode, encoding=encoding) as f:
            f.write(data)

    if gc_enabled: gc.enable()


def __validate_io_arguments(path,
                            encoding,
                            mode,
                            filetype,
                            kwargs,
                            read_or_write,
                            is_data_bytes=False):

    path          = __validate_path(path)
    filetype      = __validate_filetype(path, filetype)
    mode          = __validate_mode(mode, filetype, is_data_bytes)
    encoding      = __validate_encoding(mode, encoding)
    read_or_write = __validate_read_or_write(path, mode, read_or_write)
    kwargs        = __validate_file_keyword_args(encoding,
                                                 filetype,
                                                 kwargs,
                                                 read_or_write)
    return (path,
            encoding,
            filetype,
            mode,
            kwargs)


def __validate_path(path):
    return standardize_path(path, abspath=True)


def __validate_filetype(path, filetype):
    """ check for file types that require specialized io libraries / protocols """
    notimplemented_extensions = {'.7z',
                                 '.zip',
                                 '.gzip',
                                 '.bz',
                                 '.tar',
                                 '.png',
                                 '.jpg',
                                 '.jpeg',
                                 '.gif',
                                 '.pdf'}

    filetype = filetype or parse_file_extension(path, include_dot=True)
    filetype = filetype.lower()

    if not filetype.startswith('.'):
        filetype = '.' + filetype

    if filetype.startswith('.xl') or filetype in notimplemented_extensions:
        raise NotImplementedError("file type not supported: '{}'".format(filetype))

    return filetype


def __validate_mode(mode, filetype, is_data_bytes):
    is_rw_mode          = ('+' in mode)
    is_bytes_mode       = ('b' in mode)
    is_pickle_extension = (filetype in pickle_extensions)
    is_csv_extension    = (filetype in ('.csv',))

    if is_rw_mode:
        raise ValueError('read-write mode not supported')

    if is_csv_extension and is_bytes_mode:
        raise ValueError('csv filetype does not accept bytes mode')

    if is_pickle_extension and (not is_bytes_mode):
        mode = mode + 'b'
    elif is_data_bytes and (not is_bytes_mode):
        mode = mode + 'b'

    return mode


def __validate_encoding(mode, encoding):
    is_bytes_mode = ('b' in mode)

    if is_bytes_mode:
        encoding = None

    return encoding


def __validate_file_keyword_args(encoding, filetype, kwargs, read_or_write):
    kwargs = kwargs or {}

    if 'kwargs' in kwargs:
        kwargs.update(kwargs.pop('kwargs'))

    if filetype == '.csv' and read_or_write == 'read':
        kwargs['newline'] = kwargs.get('newline', '')
        kwargs['nrows']   = kwargs.get('nrows',   None)

    elif filetype == '.csv' and read_or_write == 'write':
        kwargs['newline'] = kwargs.get('newline', '')

    elif filetype == '.json' and read_or_write == 'write':
        kwargs['indent']       = kwargs.get('indent',       4)
        kwargs['ensure_ascii'] = kwargs.get('ensure_ascii', encoding in (None, 'ascii'))

        if ultrajson_installed:
            for k in ('default', 'separators'):
                if k in kwargs: del kwargs[k]
        else:
            kwargs['default'] = kwargs.get('default', json_unhandled_conversion)

    return kwargs


# noinspection PyRedundantParentheses
def __validate_read_or_write(path, mode, read_or_write):
    is_read_mode   = ('r' in mode)
    is_write_mode  = ('w' in mode)
    is_append_mode = ('a' in mode)
    is_url         = is_path_a_url(path)

    if read_or_write.startswith('r'):
        if (not is_read_mode) or is_append_mode:
            raise ValueError('invalid read mode: {}'.format(mode))

    elif read_or_write.startswith('w'):
        if (not is_write_mode):
            raise ValueError('invalid write mode: {}'.format(mode))
        if is_url:
            raise NotImplementedError('can not write to url: {}'.format(path))

    return read_or_write


def __read_csv(path, mode, encoding, kwargs):
    """
    _csv.Error: new-line character seen in unquoted field - do you need to open the file in universal-newline mode?
        fixed by passing lineterminator='\r'
        or newline='\r', newline=''
    """
    # region {closure}
    def read_csv_rows():
        if read_all_rows:
            return list(csv_reader)

        m = []
        try:
            for i in range(nrows):
                m.append(next(csv_reader))
        except StopIteration:
            pass

        return m

    # noinspection PyUnusedLocal
    def remove_url_from_rows(m):
        num_deleted = 0

        try:
            while is_path_a_url(m[0][0]):
                del m[0]
                num_deleted += 1
        except Exception as e:
            pass

        # if nrows specified, append rows to make up for deleted rows
        if not read_all_rows:
            try:
                for _ in range(num_deleted):
                    m.append(next(csv_reader))
            except StopIteration:
                pass

        return m
    # endregion

    newline = kwargs.pop('newline')
    nrows   = kwargs.pop('nrows')
    read_all_rows = (nrows is None)

    if is_path_a_url(path):
        with StringIO(__url_request(path, encoding=encoding), newline=newline) as f:
            csv_reader = csv.reader(f, **kwargs)
            csv_m = read_csv_rows()
            csv_m = remove_url_from_rows(csv_m)

            return csv_m

    with open(path, mode, encoding=encoding, newline=newline) as f:
        csv_reader = csv.reader(f, **kwargs)
        csv_m = read_csv_rows()

        return csv_m


def __read_json(path, mode, encoding, kwargs):
    if is_path_a_url(path):
        s = __url_request(path, encoding=encoding)
        return json.loads(s, **kwargs)

    with open(path, mode, encoding=encoding) as f:
        return json.load(f, **kwargs)


def __url_request(url, encoding=None):
    with urlopen(url) as request:
        if request.code != 200:
            raise IOError('bad url request: ({}) {}'.format(request.code, url))

        byte_string = request.read()

        if encoding:
            return byte_string.decode(encoding)
        else:
            return byte_string


def json_dumps_extended(o, **kwargs):
    kwargs['ensure_ascii'] = kwargs.get('ensure_ascii', False)
    kwargs['indent']       = kwargs.get('indent',       4)
    kwargs['default']      = kwargs.get('default',      json_unhandled_conversion)

    if ultrajson_installed:
        del kwargs['default']

    return json.dumps(o, **kwargs)


def json_unhandled_conversion(v):
    """
    convert certain python objects to json string representations, eg:
        date, datetime, set
    """
    if isinstance(v, date):
        return v.isoformat()
    if isinstance(v, set):
        return list(v)
    # if isinstance(v, Decimal) ?

    raise TypeError("object '{}', of type '{}' is not JSON serializable".format(v, type(v)))


def clear_dir(filedir):
    filedir = standardize_path(filedir, abspath=True)

    if not os.path.exists(filedir):
        return

    if not os_isdir(filedir):
        raise TypeError('"{}" is not a directory'.format(filedir))

    for item in os.listdir(filedir):
        path = filedir + item

        if os_isdir(path):
            shutil.rmtree(path)
        else:
            os.remove(path)


def copy_dir(s_dir, d_dir, ignore=None, dirs_exist_ok=False):
    """
    shutil.copytree will fail if shutil.rmtree hasn't finished deleting folder:
        PermissionError: [WinError 5] Access is denied
    """
    s_dir = standardize_path(s_dir)
    d_dir = standardize_path(d_dir)

    if ignore:
        if isinstance(ignore, str):
            ignore = (ignore,)
        elif isinstance(ignore, set):
            ignore = tuple(ignore)

        ignore = shutil.ignore_patterns(*ignore)

    if not dirs_exist_ok and os.path.exists(d_dir):
        shutil.rmtree(d_dir)

    shutil.copytree(src=s_dir,
                    dst=d_dir,
                    ignore=ignore,
                    dirs_exist_ok=dirs_exist_ok)


def file_creation_datetime(path):
    unix_t = os.path.getctime(path)
    return datetime.fromtimestamp(unix_t)


def file_modified_datetime(path):
    unix_t = os.stat(path).st_mtime
    return datetime.fromtimestamp(unix_t)


@lru_cache(maxsize=8)
def is_path_a_url(path):
    u = urlparse(path).netloc
    return bool(u)


def parse_path(path,
               pathsep='/',
               abspath=False) -> namedtuple:

    ParsedPath = namedtuple('ParsedPath', ('directory',
                                           'filename',
                                           'extension'))

    _path_ = path
    _path_ = standardize_path(_path_, pathsep, abspath)

    directory, filename  = os.path.split(_path_)
    filename,  extension = os.path.splitext(filename)

    directory = standardize_path(directory, pathsep, abspath)

    return ParsedPath(directory, filename, extension)


def parse_file_name(path):
    filename, _ = os.path.splitext(path)

    return filename


def parse_file_extension(filename, include_dot=True):
    _, extension = os.path.splitext(filename)
    extension    = extension.strip()

    has_dot = extension.startswith('.')

    if (include_dot is False) and has_dot:
        extension = extension[1:]
    elif include_dot and (not has_dot):
        extension = '.' + extension

    return extension


def standardize_path(path,
                     pathsep='/',
                     abspath=False):

    _path_ = path or ''
    if not _path_ and (not abspath):
        return ''

    if (not abspath) and (not os.path.isabs(_path_)):
        _path_ = os.path.relpath(_path_)
    else:
        _path_ = os.path.realpath(_path_)

    if pathsep == '/':
        _path_ = _path_.replace('\\', '/')
    else:
        _path_ = _path_.replace('/', '\\')

    _path_ = _path_.replace('"', '') \
                   .replace("'", '')

    if _path_ and os_isdir(_path_):
        if not _path_.endswith(pathsep):
            _path_ += pathsep

    return _path_


def traverse_dir(rootdir='.',
                 pathsep='/',
                 abspath=True,
                 *,
                 recurse=False,
                 subdirs_only=False,
                 files_only=False):

    if not is_windows_os:
        raise NotImplementedError('only available on Windows OS')

    if (subdirs_only is False) and (files_only is False):
        subdirs_only = True
        files_only   = True

    rootdir = standardize_path(rootdir, pathsep, abspath)

    if recurse:
        args = 'dir /b /s "{}"'.format(rootdir)
    else:
        args = 'dir /b "{}"'.format(rootdir)

    with os.popen(args) as os_cmd:
        out = os_cmd.read()
        out = out.split('\n')
        out.sort()

    if out and out[0] == '':
        del out[0]

    if recurse is False:
        out = [rootdir + p for p in out]

    if subdirs_only and files_only:
        return out
    elif subdirs_only:
        return [path for path in out
                     if os_isdir(path)]
    elif files_only:
        return [path for path in out
                     if os_isfile(path)]


def validate_path_exists(path):

    p_path = parse_path(path, abspath=True)
    path   = ''.join(p_path)

    if not os.path.exists(p_path.directory):
        raise FileNotFoundError('directory not found: \n\t{}'.format(p_path.directory))

    if not os.path.exists(path):
        glob_paths = glob(p_path.directory + p_path.filename + '.*')
        if glob_paths:
            extension = parse_file_extension(os.path.split(glob_paths[0])[1], include_dot=True)
            raise FileNotFoundError('file extension not found: '
                                    '\n\t{}'
                                    '\n\t{}'
                                    '\n\tdid you mean: {}?'.format(p_path.directory,
                                                                   p_path.filename,
                                                                   extension))

        raise FileNotFoundError('file not found: '
                                '\n\t{}'
                                '\n\t{}'.format(p_path.directory,
                                                p_path.filename + p_path.extension))

    return path




from collections import namedtuple

from typing import Dict
from typing import List
from typing import Union

from ..util.classes.namespace_cls import namespace_cls
from ..util.iter import is_header_row

from ..util.text import surround_double_brackets
from ..util.text import surround_single_brackets
from ..util.text import format_integer

from ..conditional import ordereddict
from ..conditional import numpy_installed

if numpy_installed:
    import numpy


class flux_row_cls:

    @classmethod
    def reserved_names(cls):
        return sorted(['headers', 'values', 'row_label'] + dir(cls),
                      reverse=True)

    def __init__(self, headers, values, row_label=None):
        """
        :param headers: OrderedDict of {'header': int}
            headers is a single dictionary passed byref from the flux_cls to many flux_row_cls instances
            this eliminates need for all flux_row_cls objects to maintain a seprate copy of these mappings and
            allows for centralized and instantaneous updatdes
        :param values: list of underlying data

        properties must be set on self.__dict__ instead of directly on self to prevent
        premature __setattr__ lookups
        """
        ''' @types '''
        self.headers:   Dict[Union[str, bytes], int]
        self.values:    List
        self.row_label: Union[int, str]

        self.__dict__['headers']   = headers
        self.__dict__['values']    = values
        self.__dict__['row_label'] = row_label

    @property
    def _preview_as_tuple(self) -> List:
        """ to help with debugging """
        names  = [surround_double_brackets(n) for n in self.header_names()]
        values = list(self.__dict__['values'])

        c_m = max(len(names), len(values))
        names.extend(['🗲missing🗲']  * (c_m - len(names)))
        values.extend(['🗲missing🗲'] * (c_m - len(values)))

        label = self.__dict__['row_label']
        if isinstance(label, int):
            names.insert(0,  surround_single_brackets('label'))
            values.insert(0, surround_single_brackets(format_integer(label)))
        elif isinstance(label, str):
            names.insert(0,  surround_single_brackets('label'))
            values.insert(0, surround_single_brackets(label))

        m = list(zip(names, values))

        return m

    @property
    def _preview_as_array(self):
        """
        to help with debugging: meant to trigger a debugging feature in PyCharm
        PyCharm will recognize the ndarray and enable the "...view as array"
        option in the debugger which displays values in a special window as a table
        """
        m = self._preview_as_tuple
        if numpy_installed:
            m = numpy.array(m, dtype=object)

        return m

    def header_names(self):
        return list(self.__dict__['headers'].keys())

    def is_jagged(self):
        return len(self.__dict__['headers']) != len(self.__dict__['values'])

    def is_empty(self):
        return (len(self.__dict__['headers']) == 0) and (len(self.__dict__['values']) == 0)

    def is_header_row(self):
        """ determine if underlying values match self.headers.keys

        self.names == self.values will not always work, since map_values_to_enum()
        was used to modify self.headers values into more suitable dictionary keys,
        like modifying duplicate values to ensure they are unique, etc
        """
        return is_header_row(self.__dict__['values'],
                             self.__dict__['headers'])

    def dict(self):
        return ordereddict(zip(self.header_names(), self.__dict__['values']))

    def namedrow(self):
        return namespace_cls(zip(self.header_names(), self.__dict__['values']))

    def namedtuple(self):
        row_nt = namedtuple('Row', self.header_names())
        return row_nt(*self.__dict__['values'])

    def join_values(self, other, names=None):
        """ copies all values where header names are shared with row_b
        :type other: flux_row_cls
        :type names: list[str]
        """
        ''' @types '''
        other: flux_row_cls
        names: Union[str, List]

        if not isinstance(other, flux_row_cls):
            raise TypeError('row expected to be flux_row_cls')

        headers_a = self.__dict__['headers']
        values_a  = self.__dict__['values']

        headers_b = other.__dict__['headers']
        values_b  = other.__dict__['values']

        _names_ = names or (headers_a.keys() & headers_b.keys())
        if not _names_:
            raise ValueError('no intersecting column names')

        if isinstance(_names_, str):
            _names_ = [_names_]

        for name in _names_:
            i_a = headers_a[name]
            i_b = headers_b[name]

            values_a[i_a] = values_b[i_b]

    def __getattr__(self, name):
        """  eg:
             o = row.column
        """
        try:
            i = self.__dict__['headers'].get(name, name)
            return self.__dict__['values'][i]
        except (TypeError, IndexError):
            self.__raise_attribute_error(name, self.headers)

    def __setattr__(self, name, value):
        """ eg:
            row.column = o
        """
        try:
            i = self.__dict__['headers'].get(name, name)
            self.__dict__['values'][i] = value
        except (TypeError, IndexError) as e:

            if isinstance(self.__dict__['values'], tuple):
                raise e
            elif name in self.__dict__:
                self.__dict__[name] = value
            else:
                self.__raise_attribute_error(name, self.headers)

    def __getitem__(self, name):
        """ eg:
            o = row['column']
        """
        try:
            i = self.__dict__['headers'].get(name, name)
            return self.__dict__['values'][i]
        except (TypeError, IndexError):
            if isinstance(name, slice):
                return self.__dict__['values'][name]

            self.__raise_attribute_error(name, self.headers)

    def __setitem__(self, name, value):
        """ eg:
            row['column'] = o
        """
        try:
            i = self.__dict__['headers'].get(name, name)
            self.__dict__['values'][i] = value
        except (TypeError, IndexError) as e:

            if isinstance(self.__dict__['values'], tuple):
                raise e
            elif isinstance(name, slice):
                self.__dict__['values'][name] = value
            elif name in self.__dict__:
                self.__dict__[name] = value
            else:
                self.__raise_attribute_error(name, self.headers)

    def __len__(self):
        return len(self.__dict__['values'])

    def __bool__(self):
        return bool(self.__dict__['values'])

    def __iter__(self):
        return iter(self.__dict__['values'])

    def __eq__(self, other):
        a = id(self.__dict__['headers'])  + hash(tuple(self.__dict__['values']))
        b = id(other.__dict__['headers']) + hash(tuple(other.__dict__['values']))

        return a == b

    def __ne__(self, other):
        return not self.__eq__(other)

    def __repr__(self):
        label = self.__dict__['row_label']
        if isinstance(label, int):
            row_label = format_integer(label)
            row_label = surround_single_brackets(row_label)
            row_label = '{} '.format(row_label)
        elif isinstance(label, str):
            row_label = surround_single_brackets(label)
            row_label = '{} '.format(row_label)
        else:
            row_label = ''

        if self.is_jagged():
            jagged_label = len(self.__dict__['values']) - len(self.__dict__['headers'])
            jagged_label = surround_single_brackets('{:+}'.format(jagged_label))
            jagged_label = '🗲jagged {}🗲  '.format(jagged_label)
        else:
            jagged_label = ''

        if self.is_header_row():
            values = ', '.join([str(n) for n in self.header_names()])
            values = surround_double_brackets(values)
        else:
            values = (repr(self.__dict__['values']).replace('"', '')
                                                   .replace("'", ''))

        return '{}{}{}'.format(row_label, jagged_label, values)

    @staticmethod
    def __raise_attribute_error(invalid, headers):
        indices = [str(i) for i in headers.values()]
        _nf_ = max(len(n) for n in indices)
        _nf_ = '{: <%s}' % str(_nf_)
        indices = [_nf_.format(n) for n in indices]

        s = ["{}: '{}'".format(i, n) for i, n in zip(indices, headers.keys())]
        s = '\n\t'.join(s)

        s = ("'{}' column name does not exist in columns: "
             "\n\t{}".format(invalid, s))

        raise AttributeError(s) from None





import os
import sys

''' 
config: 
    default settings for vengeance.util functions
    eg, vengeance/config.ini:
        [console]
        effect = bold
        enable_ansi_escape = True
        
        [filesystem]
        encoding = utf-8-sig

loads_excel_module: 
    determines if excel_com module should be loaded in vengeance.__init__ 
    if environment is expected to support Windows COM interface, load vengeance.excel_com module
    
    vengeance.excel_com functions:
        vengeance.open_workbook
        vengeance.close_workbook
        vengeance.lev_cls
        etc ...

ordereddict:
    starting at python 3.6, the built-in dict is both insertion-ordered and compact, 
    using about half the memory of collections.OrderedDict 
'''
config             = {}
config_file_loaded = False

python_version      = sys.version_info
is_windows_os       = (os.name == 'nt' or sys.platform == 'win32')
is_utf_console      = ('utf' in sys.stdout.encoding.lower())
is_pypy_interpreter = ('__pypy__' in set(sys.builtin_module_names))
loads_excel_module  = (not is_pypy_interpreter and is_windows_os)

ordereddict             = dict
dateutil_installed      = False
ultrajson_installed     = False
numpy_installed         = False
line_profiler_installed = False


if python_version < (3, 6):
    from collections import OrderedDict
    ordereddict = OrderedDict

if python_version < (3, 9):
    try:
        import ujson
        ultrajson_installed = True
    except ImportError:
        pass

try:
    import dateutil
    dateutil_installed = True
except ImportError:
    pass

try:
    import numpy
    numpy_installed = True
except ImportError:
    pass

try:
    import line_profiler
    line_profiler_installed = True
except ImportError:
    pass


def load_vengeance_configuration_file():
    """
    sys.getfilesystemencoding()
    locale.getpreferredencoding()

    if cp.has_section('filesystem'):
        cp_section = cp['filesystem']
        config.update({'encoding': cp_section.get('encoding')})
    """
    global config
    global config_file_loaded

    config_file_loaded = True

    vcf = __load_vengeance_configuration_file()
    if vcf is None:
        return

    if vcf.has_section('console'):
        cp_section = vcf['console']

        config.update({'color':              cp_section.get('color'),
                       'effect':             cp_section.get('effect'),
                       'formatter':          cp_section.get('formatter'),
                       'enable_ansi_escape': cp_section.getboolean('enable_ansi_escape')})
        if config['enable_ansi_escape']:
            __enable_ansi_escape_in_windows_console()


def __load_vengeance_configuration_file():
    import site

    config_paths = [site.getsitepackages()[1]  + '\\vengeance\\config.ini']

    if is_windows_os:
        try:    config_paths.append(os.environ['localappdata'] + '\\Temp\\vengeance\\config.ini')
        except: pass

    for config_path in config_paths:
        if os.path.exists(config_path):
            from configparser import ConfigParser

            vcf = ConfigParser()
            vcf.read(config_path)

            return vcf

    return None


def __enable_ansi_escape_in_windows_console():
    """ attempt to turn on ansi escape colors in console

    ansi escape: \x1b, \033, <0x1b>, 
    how to effectively check if console supports ansi escapes?
        ansi escape shows up as '←' in python.exe console

    # os.popen('chcp 65001')

    import codecs
    import locale
    a = codecs.getwriter(locale.getpreferredencoding())(sys.stdout)

    STD_OUTPUT_HANDLE = -11
    DWMODE            = 7

    # GetConsoleMode = ctypes.windll.kernel32.GetConsoleMode
    # print(f'{GetConsoleMode(STD_OUTPUT_HANDLE) = }')
    """
    if not is_windows_os:
        return

    try:
        import ctypes
        global is_utf_console

        SetConsoleMode    = ctypes.windll.kernel32.SetConsoleMode
        GetStdHandle      = ctypes.windll.kernel32.GetStdHandle
        STD_OUTPUT_HANDLE = -11
        DWMODE            = 7

        SetConsoleMode(GetStdHandle(STD_OUTPUT_HANDLE), DWMODE)
        is_utf_console = True

    except:
        is_utf_console = False


# config_file_loaded = True
if config_file_loaded is False:
    load_vengeance_configuration_file()


# ordereddict             = dict
# is_pypy_interpreter     = True
# is_windows_os           = True
# dateutil_installed      = False
# ultrajson_installed     = False
# numpy_installed         = False
# line_profiler_installed = False
# loads_excel_module      = False




from .lev_cls import lev_cls

# following more traditional PascalCase naming convention for classes
# noinspection PyPep8Naming
from .lev_cls import lev_cls as Lev

__all__ = ['lev_cls',
           'Lev']




import ctypes
import gc
import os
import pythoncom

# noinspection PyUnresolvedReferences
from pythoncom import com_error as pythoncom_error
from _ctypes import COMError as ctypes_error

from ctypes import byref
from ctypes import c_void_p
from ctypes import py_object
from ctypes import POINTER
from ctypes import PyDLL
from ctypes.wintypes import BOOL

from comtypes import IUnknown
from comtypes.client          import CreateObject as comtypes_createobject
from comtypes.automation      import IDispatch    as comtypes_idispatch
from comtypes.client.dynamic  import Dispatch     as comtypes_dispatch

from time import sleep

from win32com.client.gencache import EnsureDispatch                          # early-bound references

from ..util.filesystem import parse_path
from ..util.filesystem import standardize_path
from ..util.filesystem import validate_path_exists

from ..util.text import styled
from ..util.text import vengeance_message
from ..util.text import vengeance_warning

from .excel_constants import (xlMaximized,
                              xlNormal,
                              xlMinimized)

# Windows api functions
FindWindowExA              = ctypes.windll.user32.FindWindowExA
SetForegroundWindow        = ctypes.windll.user32.SetForegroundWindow
AccessibleObjectFromWindow = ctypes.oledll.oleacc.AccessibleObjectFromWindow
OpenProcess                = ctypes.windll.kernel32.OpenProcess
TerminateProcess           = ctypes.windll.kernel32.TerminateProcess
CloseHandle                = ctypes.windll.kernel32.CloseHandle

corrupt_hwnds = set()

NATIVE_OM         = -16
WM_CLOSE          = 16
PROCESS_TERMINATE = 1


def open_workbook(path,
                  excel_app='new',
                  *,
                  reload_addins=False,
                  display_alerts=False,
                  enable_events=True,
                  windowstate=xlMaximized,
                  **kwargs):

    wb = get_opened_workbook(path)
    if wb is None:
        wb = __open_workbook_dispatch(path,
                                      excel_app,
                                      reload_addins=reload_addins,
                                      display_alerts=display_alerts,
                                      enable_events=enable_events,
                                      windowstate=windowstate,
                                      **kwargs)

    kw_read_only = kwargs.get('ReadOnly', False)
    wb_read_only = wb.ReadOnly

    if (wb_read_only is False) and kw_read_only:
        raise ValueError("'{}' NOT open as read-only".format(wb.Name))
    elif wb_read_only:
        print(vengeance_message("'{}' open as read-only".format(wb.Name)))

    return wb


def get_opened_workbook(path):
    """ scan all open workbooks across all Excel instances for matching paths

    :return captured Workbook or None
    """
    p_path = parse_path(path.lower(), abspath=False)
    name   = p_path.filename + p_path.extension

    workbooks = []
    for excel_app in all_excel_instances():
        for wb in excel_app.Workbooks:

            wb_path = wb.FullName
            wb_name = wb.Name.lower()

            if (p_path.directory and os.path.samefile(path, wb_path)) or \
               (name == wb_name):

                workbooks.append(wb)

    if not workbooks:
        return None

    if len(workbooks) > 1:
        vengeance_warning("multiple workbooks in different Excel instances found: '{}' "
                          .format([wb.FullName for wb in workbooks]))

    return workbooks[0]


def is_workbook_open(path):
    wb = get_opened_workbook(path)
    return wb is not None


def close_workbook(wb, save):
    """
    provides a few conveniences that the client may not want to deal with
        1) if Workbook is the only one in the Excel application, closes the application as well
        2) correctly releases the com pointer for the workbook
    """
    import win32gui
    import win32process

    # from win32gui import PostMessage
    # from win32api import OpenProcess
    # from win32api import TerminateProcess
    # from win32api import CloseHandle
    # GetWindowThreadProcessId   = ctypes.windll.user32.GetWindowThreadProcessId

    if save and wb.ReadOnly:
        raise AssertionError("'{}' is open as read-only, cannot save and close".format(wb.Name))

    excel_app      = wb.Application
    window_h       = excel_app.Hwnd
    display_alerts = excel_app.DisplayAlerts

    excel_app.DisplayAlerts = False
    wb.Close(save)
    excel_app.DisplayAlerts = display_alerts

    __close_blank_workbooks(excel_app)

    if __is_excel_application_empty(excel_app):
        excel_app.Quit()
        sleep(1)

        tid, pid = win32process.GetWindowThreadProcessId(window_h)
        win32gui.PostMessage(window_h, WM_CLOSE, 0, 0)
        sleep(1)

        try:
            hid = OpenProcess(PROCESS_TERMINATE, 0, pid)
            if hid:
                TerminateProcess(hid, 0)
                CloseHandle(hid)
                sleep(1)

        except Exception:
            pass

    del excel_app
    del wb

    gc.collect()


def __open_workbook_dispatch(path,
                             excel_app,
                             *,
                             reload_addins,
                             display_alerts,
                             enable_events,
                             windowstate,
                             **kwargs):
    """
    UpdateLinks
        0: don’t update
        1: update external links but not remote links
        2: update remote links but not external links
        3: update all links
        default: prompt the user
    """
    if windowstate not in (None, xlMaximized, xlNormal, xlMinimized):
        raise ValueError('windowstate must be in (None, xlMaximized ({}), xlNormal ({}), xlMinimized ({}))'
                         .format(xlMaximized, xlNormal, xlMinimized))

    path      = validate_path_exists(path)
    excel_app = __validate_excel_dispatch(excel_app)

    if display_alerts:
        excel_application_to_foreground(excel_app, windowstate, add_workbook_if_empty=False)

    _display_alerts_ = excel_app.DisplayAlerts
    _enable_events_  = excel_app.EnableEvents

    excel_app.DisplayAlerts = display_alerts
    excel_app.EnableEvents  = enable_events
    wb = excel_app.Workbooks.Open(path,
                                  **kwargs)
    excel_app.DisplayAlerts = _display_alerts_
    excel_app.EnableEvents  = _enable_events_

    excel_application_to_foreground(excel_app, windowstate, add_workbook_if_empty=False)

    if reload_addins:
        reload_all_add_ins(excel_app)

    return wb


def __validate_excel_dispatch(excel_app):
    if excel_app in (None, 'new'):   return new_excel_application()
    elif excel_app == 'any':         return any_excel_application()
    elif excel_app == 'empty':       return empty_excel_application()

    if not hasattr(excel_app, 'Workbooks'):
        raise ValueError("excel_app parameter must be in (None, 'new', 'any', 'empty') or an Excel application pointer")

    return excel_app


def new_excel_application():
    """
    ?
    excel_app = comtypes_createobject('Excel.Application', dynamic=False)
    """
    excel_app = comtypes_createobject('Excel.Application', dynamic=True)
    excel_app = __iunknown_pointer_to_python_object(excel_app, comtypes_idispatch)
    excel_app = EnsureDispatch(excel_app)

    return excel_app


def any_excel_application():
    excel_app = EnsureDispatch('Excel.Application') or \
                new_excel_application()
    return excel_app


def empty_excel_application():

    for excel_app in all_excel_instances():
        if __is_excel_application_empty(excel_app):
            print(vengeance_message('utilizing empty Excel instance'))
            return excel_app

    raise AssertionError('empty Excel application not found')


def __is_excel_application_empty(excel_app):
    """ an Excel application with no workbooks or a an Excel application
    with the default workbook opened
    """
    for wb in excel_app.Workbooks:
        if not __is_workbook_blank(wb):
            return False

    return True


# noinspection PyUnusedLocal
def __close_blank_workbooks(excel_app):
    workbooks = [wb for wb in excel_app.Workbooks
                    if __is_workbook_blank(wb)]

    _display_alerts_ = excel_app.DisplayAlerts
    excel_app.DisplayAlerts = False

    for wb in workbooks:
        wb.Close(False)

        wb = None
        del wb

    excel_app.DisplayAlerts = _display_alerts_


def __is_workbook_blank(wb):
    if not wb.Saved:
        return False

    for ws in wb.Sheets:
        if ws.UsedRange.Address != '$A$1':
            return False

    return True


def all_excel_instances():
    window_h = __next_window_handle()

    excel_hwnds = set()
    while window_h != 0:
        excel_app = __excel_application_from_window_handle(window_h)

        if excel_app is None:
            window_h = __next_window_handle(window_h)
            continue

        hwnd = excel_app.Hwnd
        if hwnd in excel_hwnds:
            window_h = __next_window_handle(window_h)
            continue

        excel_hwnds.add(hwnd)
        yield excel_app

        window_h = __next_window_handle(window_h)


def reload_all_add_ins(excel_app):
    print(vengeance_message('reloading Excel add-ins'))

    for add_in in excel_app.AddIns:
        if add_in.Installed:
            name = add_in.Name

            try:
                add_in.Installed = False
                add_in.Installed = True
                print('        * {}'.format(name))
            except (ctypes_error, pythoncom_error, NameError) as e:
                vengeance_warning('    failed to load Excel add-in: {}, {}'.format(name, e))

    print()


def excel_application_to_foreground(excel_app, windowstate=None, add_workbook_if_empty=False):
    excel_app.Visible = True

    # excel_app will stay invisible until a workbook is added
    if add_workbook_if_empty and (excel_app.Visible is False) and (excel_app.Workbooks.Count == 0):
        excel_app.Workbooks.Add()
        excel_app.Visible = True

    # make sure window is not minimized (xlMinimized has weird behavior)
    if excel_app.WindowState != xlMaximized:
        excel_app.WindowState = xlNormal

    if windowstate is not None:
        excel_app.WindowState = windowstate

    SetForegroundWindow(excel_app.Hwnd)


def __next_window_handle(window_h=0):
    return FindWindowExA(0, window_h, b'XLMAIN', 0)


# noinspection PyTypeChecker,PyProtectedMember,PyUnresolvedReferences
def __excel_application_from_window_handle(window_h):
    """
    comtypes library is used to search windows handles for Excel application,
    then the application pointer is converted to a pywin object

    obj_ptr = POINTER(comtypes_idispatch)()
        Expected type [_CT], got Type[IDispatch] instead
        this is probably not correct way to set up this pointer

        from pywintypes import IID?
    """
    global corrupt_hwnds

    if window_h in corrupt_hwnds:
        return None

    xl_desk_hwnd = FindWindowExA(window_h,     0, b'XLDESK', 0)
    excel7_hwnd  = FindWindowExA(xl_desk_hwnd, 0, b'EXCEL7', 0)

    if excel7_hwnd == 0:
        corrupt_hwnds.add(window_h)
        return None

    obj_ptr = POINTER(comtypes_idispatch)()
    AccessibleObjectFromWindow(excel7_hwnd,
                               NATIVE_OM,
                               byref(obj_ptr._iid_),
                               byref(obj_ptr))

    try:
        com_ptr   = comtypes_dispatch(obj_ptr).Application
        excel_app = __iunknown_pointer_to_python_object(com_ptr, comtypes_idispatch)
    except (ctypes_error, pythoncom_error, NameError):

        # comtypes Dispatch() call rejected: user may be editing cell
        raise ChildProcessError('Remote Procedure Call to Excel Application rejected. '
                                '\n'
                                '\n\t(Excel will reject automation calls while waiting on user input, '
                                '\n\tcheck if cursor is still active within a cell somewhere)') from None

    try:
        return EnsureDispatch(excel_app)
    except AttributeError:

        # win32com Dispatch() call rejected: COM files may be corrupted
        __move_win32com_gencache_folder()

        raise ChildProcessError('Error dispatching Excel Application from win32com module. '
                                '\n'
                                '\n\t(contents of the win32com gen_py folder may be corrupt)') from None


# noinspection PyTypeChecker, PyProtectedMember
def __iunknown_pointer_to_python_object(com_ptr, interface):
    args = (POINTER(IUnknown), c_void_p, BOOL)

    object_pointer          = PyDLL(pythoncom.__file__).PyCom_PyObjectFromIUnknown
    object_pointer.restype  = py_object
    object_pointer.argtypes = args

    o = object_pointer(com_ptr._comobj,
                       byref(interface._iid_),
                       True)

    return o


# noinspection PyBroadException
def __move_win32com_gencache_folder():
    """
    move win32com gen_py cache files to site-packages folder
    from
        %userprofile%/Local/Temp/gen_py/
    to
        %python_folder%/Lib/site-packages/win32com/gen_py/

    helps prevent win32com EnsureDispatch() call rejection due to corrupted COM files
    """
    try:
        import shutil
        import site
        import subprocess
        import win32com
        from pathlib import Path

        appdata_gcf = os.environ['userprofile'] + '\\AppData\\Local\\Temp\\gen_py\\'
        site_gcf    = site.getsitepackages()[1] + '\\win32com\\gen_py\\'

        if not os.path.exists(appdata_gcf):
            appdata_gcf = win32com.__gen_path__
            if standardize_path(appdata_gcf).lower() == standardize_path(site_gcf).lower():
                return

        s = vengeance_message('Attempting to reset win32com gencache folder ...\n')
        s = styled(s, 'yellow', 'bold')
        print(s)

        if os.path.exists(appdata_gcf):
            subprocess.call('explorer.exe "{}"'.format(Path(appdata_gcf).parent))
            shutil.rmtree(appdata_gcf)

        if not os.path.exists(site_gcf):
            subprocess.call('explorer.exe "{}"'.format(Path(site_gcf).parent))
            os.makedirs(site_gcf)

    except Exception:
        pass




import re
from collections import namedtuple
from typing import Generator
from typing import Union
from typing import Any
from typing import Dict
from typing import List
from typing import Tuple
from typing import Iterable
from typing import ItemsView
from typing import KeysView
from typing import ValuesView

from ..conditional import ordereddict
from .classes.tree_cls import tree_cls
from .classes.namespace_cls import namespace_cls


vengeance_cls_names = {'flux_cls',
                       'lev_cls',
                       'excel_levity_cls'}


class IterationDepthError(TypeError):
    pass


class ColumnNameError(ValueError):
    pass


def standardize_variable_arity_values(values,
                                      depth=None,
                                      depth_offset=None):
    """
    this function is heavily utilized to correctly un-nest function arguments
    in flux_cls methods

    eg flux_cls:
        def append_columns(self, *names):
            names = standardize_variable_arity_values(names, depth=1)

    # while is_descendable(values):
    #     values = values[0]
    """

    # convert iterators, even if they are nested
    if is_descendable(values) and is_exhaustable(values[0]):
        values = iterator_to_collection(values[0])
    else:
        values = iterator_to_collection(values)

    values = modify_iteration_depth(values, depth, depth_offset,
                                    first_element_only=True)

    return values


def iteration_depth(values, first_element_only=False):
    """
    this function is heavily utilized to correctly un-nest function arguments
    in flux_cls methods

    eg:
        0 = iteration_depth('abc')
        1 = iteration_depth(['abc'])
        2 = iteration_depth([[2], [2]])

        items = [1, [2, 2, [3, 3, 3, [4, [5, 5, 5, 5], 4]]]]
        1 = iteration_depth(items, first_element_only=True)
        5 = iteration_depth(items, first_element_only=False)
    """
    if is_exhaustable(values):
        raise TypeError('cannot evaluate an exhaustable iterator')

    if isinstance(values, dict):
        values = tuple(values.values())

    if not is_collection(values):
        return 0

    if len(values) == 0:
        return 1

    if first_element_only:
        return 1 + iteration_depth(values[0], first_element_only)
    else:
        return 1 + max([iteration_depth(v, first_element_only) for v in values])


def modify_iteration_depth(values,
                           depth=None,
                           depth_offset=None,
                           first_element_only=False):
    """
    this function is heavily utilized to correctly un-nest function arguments
    in flux_cls methods

    eg:
        'a'              = modify_iteration_depth([[['a']]], depth=0)
        ['a', 'b']       = modify_iteration_depth([[['a', 'b']]], depth=0)
        [['a']]          = modify_iteration_depth('a', depth=2)
        [[[['a', 'b']]]] = modify_iteration_depth(['a', 'b'], depth=4)

        required behaviors:
            must not reduce depth of any dictionary keys or values
                {'a': None} = modify_iteration_depth([{'a': None}], depth=0)

        undecided behaviors:
            how to deal with mixed iteration depths?
                [['ab', ['cd']]]
                should this return ['ab', ['cd']] or ['ab', 'cd']?

            # has_multiple_depths = set(iteration_depth(_v_, first_element_only=False) for _v_ in v)

            if is_descendable(values):
                values = values[0]
            else:
                break

            # if first_element_only:
            #     if is_descendable(values): values = values[0]
            #     else:                      break
            # else:
            #     values = list(chain.from_iterable(values))
    """
    if depth is None and depth_offset is None:
        raise ValueError('conflicting values for depth and depth_offset')

    if isinstance(depth, int) and isinstance(depth_offset, int):
        raise ValueError('conflicting values for depth and depth_offset')

    if depth_offset is None:
        value_depth  = iteration_depth(values, first_element_only)
        depth_offset = depth - value_depth

    if depth_offset < 0:
        for _ in range(abs(depth_offset)):
            if is_descendable(values):
                values = values[0]
            else:
                break

    elif depth_offset > 0:
        for _ in range(depth_offset):
            values = [values]

    return values


def are_indices_contiguous(indices):
    if len(indices) == 1:
        return False

    for i_1, i_2 in zip(indices, indices[1:]):
        if i_2 - i_1 != 1:
            return False

    return True


def is_collection(o):
    """ determine if value is an iterable object or data structure
    function used mainly to distinguish data structures from other
    iterables

    eg:
        False = is_collection('bleh')
        True  = is_collection(['bleh'])

        False = is_collection(range(3))
        True  = is_collection(list(range(3)))
    """
    if isinstance(o, (str, bytes, range, dict, ItemsView)):
        return False
    else:
        return isinstance(o, Iterable)


# noinspection PyStatementEffect,PyBroadException
def is_subscriptable(o):
    try:
        o[0]
        return True
    except:
        return False


def is_exhaustable(o):
    """
    return (hasattr(o, '__next__') or
            isinstance(o, range))
    """
    return hasattr(o, '__next__')


def is_dictview(o):
    return (isinstance(o, KeysView) or
            isinstance(o, ValuesView) or
            isinstance(o, ItemsView))


def is_vengeance_class(o):
    base_cls_names = set(base_class_names(o))

    return bool(base_cls_names & vengeance_cls_names)


def is_descendable(o):
    return (is_collection(o) and
            is_subscriptable(o) and
            len(o) == 1)


def iterator_to_collection(o):
    if is_vengeance_class(o):
        return list(o.values())

    if isinstance(o, (range, set)) or \
       is_exhaustable(o) or \
       is_dictview(o):

        return list(o)

    return o


def base_class_names(o):
    return [b.__name__ for b in o.__class__.mro()]


def transpose(m, astype=None) -> Generator[Union[List, Tuple], None, None]:
    m = iterator_to_collection(m)
    n = iteration_depth(m, first_element_only=True)

    if n == 0:
        raise IterationDepthError('matrix must have at least 1 iterable dimension')

    if astype is None:
        if n == 1: astype = type(m)
        else:      astype = type(m[0])

    if astype is list:
        if n == 1: t = ([row] for row in m)
        else:      t = (list(row) for row in zip(*m))
    else:
        if n == 1: t = ((row,) for row in m)
        else:      t = zip(*m)

    return t


def to_grouped_dict(flat_dict: Dict) -> Dict[Any, Dict]:
    """ re-map keys (tuples) to a nested structure

    eg:
        # "flat" keys
        d = {('a₁', 'b₁', 'c₁', 'd₁'): 'v_1',
             ('a₁', 'b₂', 'c₁', 'd₂'): 'v_2',
             ('a₂', 'b₁', 'c₁', 'd₁'): 'v_4',
             ('a₂', 'b₁', 'c₁', 'd₂'): 'v_3'}

        {'a₁': {'b₁': {'c₁': {'d₁': 'v_1'}},
                'b₂': {'c₁': {'d₂': 'v_2'}}},
         'a₂': {'b₁': {'c₁': {'d₁': 'v_4',
                              'd₂': 'v_3'}}}} = to_grouped_dict(d)
    """
    nested_dict = tree_cls(flat_dict).traverse()

    return nested_dict


def is_header_row(values, headers):
    """ determine if underlying values match headers.keys

    checking headers.keys() == values will not always work, since map_values_to_enum()
    was used to modify names into more suitable header keys,
    like modifying duplicate values to ensure they are unique

    eg:
        True = is_header_row(['a', 'b', 'c'],
                             {'a': 0, 'b': 1, 'c': 2})
        True = is_header_row(['a', 'b', 'b'],
                             {'a': 0, 'b': 1, 'b__2': 2})

    """
    if not headers:
        return False

    value_names  = map_values_to_enum(values).keys()
    header_names = set(headers)

    return value_names == header_names


def map_values_to_enum(sequence, start=0, to_snake_case=False) -> Dict[Union[str, bytes], int]:
    """ :return {unique_key: i: int} for all items in sequence

    values are modified before they are added as keys:
        * values coerced to string (if not bytes)
        * non-unique keys are appended with '__{num}' suffix

    eg
        {'a':    0,
         'b':    1,
         'c':    2,
         'b__2': 3,
         'b__3': 4} = map_values_to_enum(['a', 'b', 'c', 'b', 'b'])
    """
    duplicates = {}
    values_to_indices = ordereddict()

    for i, v in enumerate(sequence, start):
        is_bytes = isinstance(v, bytes)

        if is_bytes:
            v_str = v.decode()
        else:
            v_str = str(v)

        if to_snake_case:
            v_str = snake_case(v_str)

        is_duplicate = (v_str in duplicates)

        if not is_duplicate:
            duplicates[v_str] = 1
        else:
            duplicates[v_str] += 1
            v_str = '{}__{}'.format(v_str, duplicates[v_str])

        if is_bytes:
            v_str = v_str.encode()

        values_to_indices[v_str] = i

    return values_to_indices


# noinspection PyProtectedMember
def to_namespaces(o, to_snake_case=False):
    """ recursively convert values to namespaces """

    if hasattr(o, '_fields'):
        if to_snake_case:
            fields = [snake_case(k) for k in o._fields]
        else:
            fields = o._fields

        d = [(k, to_namespaces(v)) for k, v in zip(fields, o)]
        return namespace_cls(d)

    elif isinstance(o, dict):
        if to_snake_case:
            d = [(snake_case(k), to_namespaces(v)) for k, v in o.items()]
        else:
            d = [(k, to_namespaces(v)) for k, v in o.items()]

        return namespace_cls(d)

    elif is_collection(o):
        return [to_namespaces(v) for v in o]
    else:
        return o


# noinspection PyArgumentList, PyProtectedMember
def to_namedtuples(o, to_snake_case=False):
    is_namedtuple = (isinstance(o, tuple) and
                     type(o) is not tuple)

    if is_namedtuple:
        if to_snake_case:
            fields = [snake_case(k) for k in o._fields]
        else:
            fields = o._fields

        nt = namedtuple(type(o).__name__, fields)
        return nt(*[to_namedtuples(v) for v in o])

    if isinstance(o, dict):
        if to_snake_case:
            fields = [snake_case(n) for n in o.keys()]
        else:
            fields = o.keys()

        nt = namedtuple('NT', fields)
        return nt(*[to_namedtuples(v) for v in o.values()])

    elif is_collection(o):
        return [to_namedtuples(v) for v in o]
    else:
        return o


# noinspection DuplicatedCode
def snake_case(s):
    """ eg:
        'some_value' = snake_case('someValue')

    circular dependencies?
        from .text import snake_case as _snake_case_
        return _snake_case_(s)
    """
    s: str

    camel_re = re.compile('''
        (?<=[a-z])[A-Z](?=[a-z])
    ''', re.VERBOSE)

    s = s.strip()
    if s[0].isupper():
        s = s[0].lower() + s[1:]

    for i, match in enumerate(camel_re.finditer(s)):
        c = match.span()[0] + i
        p = '_' + s[c].lower()
        s = '{}{}{}'.format(s[:c], p, s[c + 1:])

    return s.lower()





import os
import sys

from collections import deque
from collections import namedtuple

from logging import Logger
from logging import Formatter
from logging import FileHandler
from logging import StreamHandler
from logging import LogRecord
from logging import (NOTSET,
                     DEBUG,
                     INFO,
                     WARNING,
                     ERROR,
                     FATAL,
                     CRITICAL)
# from logging import getLevelName as get_level_name

from textwrap import dedent
from typing import List

from .. util.filesystem import parse_path
from .. util.filesystem import standardize_path
from .. util.text import object_name
from .. util.text import styled
from .. util.text import flush_stdout


class log_cls(Logger):
    # max_num_records = None
    max_num_records = 10

    # default_time_format = '%Y-%m-%d %H:%M:%S'
    # default_msec_format = '%s,%03d'

    levels_mapping = {'notset':   NOTSET,
                      'debug':    DEBUG,
                      'info':     INFO,
                      'warning':  WARNING,
                      'error':    ERROR,
                      'fatal':    CRITICAL,
                      'critical': CRITICAL,

                      NOTSET:     'notset',
                      DEBUG:      'debug',
                      INFO:       'info',
                      WARNING:    'warning',
                      ERROR:      'error',
                      FATAL:      'critical',
                      CRITICAL:   'critical'}

    def __init__(self, name_or_path='',
                       level='NOTSET',
                       *,
                       write_to_stream=True,
                       write_to_file=True,
                       log_format='[{asctime}] [{levelname}] {message}',
                       date_format=None,
                       file_mode='w',
                       file_encoding='utf-8',
                       exception_callback=None,
                       colored_statements=False):
        """
        :param name_or_path:
            parsed to determine name of logger
            if name_or_path includes directory or file extension, a file handler is added
        :param log_format:
            format to be applied to handlers
            eg log_format:
                '[%(asctime)s] [%(levelname)s] %(message)s'
                {asctime}; {levelname}; "{message}"
        :param exception_callback:
            function to be invoked when self.exception_handler() is called
            sys.excepthook is automatically set to self.exception_handler if
            exception_callback is a valid function

        log_format='[{levelname}] [{asctime}] {message}',
        log_format='[{asctime}] [{levelname}] [{threadName}] [{process}] {message}',
        log_format='[{asctime}.{msecs:.4f}] [{levelname}] [{threadName}] [{process}] {message}'
        date_format='%Y-%m-%d %I:%M:%S %p',

        %(name)s            Name of the logger (logging channel)
        %(levelno)s         Numeric logging level for the message (DEBUG, INFO, WARNING, ERROR, CRITICAL)
        %(levelname)s       Text logging level for the message    ('DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL')
        %(pathname)s        Full pathname of the source file where the logging call was issued (if available)
        %(filename)s        Filename portion of pathname
        %(module)s          Module (name portion of filename)
        %(lineno)d          Source line number where the logging call was issued (if available)
        %(funcName)s        Function name
        %(created)f         Time when the LogRecord was created (time.time() return value)
        %(asctime)s         Textual time when the LogRecord was created
        %(msecs)d           Millisecond portion of the creation time
        %(relativeCreated)d Time in milliseconds when the LogRecord was created, relative to the time the logging 
                            module was loaded (typically at application startup time)
        %(thread)d          Thread ID (if available)
        %(threadName)s      Thread name (if available)
        %(process)d         Process ID (if available)
        %(message)s         The result of record.getMessage(), computed just as the record is emitted
        """
        ''' @types '''
        self.records: deque[LogRecord]

        if not (write_to_stream or write_to_file):
            raise ValueError('write_to_stream or write_to_file must not both be False')

        if (exception_callback is not None) and not callable(exception_callback):
            raise TypeError('exception_callback must be callable')

        name  = parse_path(name_or_path).filename
        level = self.levelname(level)

        super().__init__(name, level)

        self.write_to_stream    = write_to_stream
        self.write_to_file      = write_to_file
        self.name_or_path       = name_or_path
        self.mode               = file_mode
        self.encoding           = file_encoding
        self.formatter          = self.log_formatter(log_format, date_format)
        self.records            = deque(maxlen=self.max_num_records)
        self.exception_callback = exception_callback
        self.exception_message  = ''

        if write_to_stream:
            self.add_stream_handler(self.level, colored_statements, sys.stdout)

        if write_to_file:
            self.add_file_handler(name_or_path, self.level, file_mode, file_encoding)

        if exception_callback:
            sys.excepthook = self.exception_handler

    @property
    def stream_handlers(self) -> List[StreamHandler]:
        handlers = []
        for h in self.handlers:
            if isinstance(h, StreamHandler) and not isinstance(h, FileHandler):
                handlers.append(h)

        return handlers

    @property
    def paths(self) -> List[str]:
        return [h.baseFilename for h in self.file_handlers]

    @property
    def paths_parsed(self) -> List[namedtuple]:
        return [parse_path(h.baseFilename) for h in self.file_handlers]

    @property
    def file_handlers(self) -> List[FileHandler]:
        handlers = []
        for h in self.handlers:
            if isinstance(h, FileHandler):
                handlers.append(h)

        return handlers

    def path_hyperlinks(self) -> str:
        links = []
        for path in self.paths:
            links.append('File "{}", line 1'.format(path))

        links = '\n'.join(links)
        return links

    def levelname(self, level=None) -> str:
        if level is None and 'level' in self.__dict__:
            if self.level is None:
                raise ValueError('self.level is None')
            else:
                return self.levelname(self.level)

        _level_ = level
        if isinstance(_level_, str):
            _level_ = _level_.lower()

        if _level_ not in self.levels_mapping:
            raise KeyError(_level_)

        if isinstance(_level_, int):
            _level_ = self.levels_mapping[_level_]

        return _level_.upper()

    def reset_formatter(self, log_format, date_format):
        self.formatter = self.log_formatter(log_format, date_format)
        for h in self.handlers:
            h.setFormatter(self.formatter)

    def reset_level(self, level):
        level = self.levelname(level).lower()
        level = self.levels_mapping[level]

        self.level = level

        for h in self.handlers:
            h.setLevel(level)

    def close(self):
        self.close_stream_handlers()
        self.close_file_handlers()

    def close_stream_handlers(self, i=None):
        handlers = self.stream_handlers
        if isinstance(i, int):
            handlers = [handlers[i]]

        for h in handlers:
            h.close()
            self.removeHandler(h)

    def close_file_handlers(self, i=None):
        handlers = self.file_handlers
        if isinstance(i, int):
            handlers = [handlers[i]]

        for h in handlers:
            h.close()
            self.removeHandler(h)

    # noinspection PyProtectedMember
    def clear_files(self, i=None):
        handlers = self.file_handlers
        if isinstance(i, int):
            handlers = [handlers[i]]

        for h in handlers:
            h.close()
            h._open()

    def remove_stream_handlers(self, i=None):
        self.close_stream_handlers(i)

    def remove_file_handlers(self, i=None):
        self.close_file_handlers(i)

    def clear_file_handlers(self, i=None):
        self.clear_files(i)

    def set_level(self, level):
        self.reset_level(level)

    def add_parent_log(self, parent_log):
        """ if trying to write to same paths?? """
        if id(parent_log) == id(self):
            raise ValueError('parent log refers to the same object as current log')

        self.parent = parent_log

    def add_stream_handler(self, level=None,
                                 colored_statements=False,
                                 stream=sys.stdout):

        if colored_statements:
            h = colored_streamhandler_cls(stream)
        else:
            h = StreamHandler(stream)

        level = self.levelname(level)

        h.setLevel(level)
        h.setFormatter(self.formatter)
        self.addHandler(h)

    def add_file_handler(self, path,
                               level=None,
                               mode='w',
                               encoding='utf-8'):

        path = self.resolve_log_path(path)
        if not path:
            self.write_to_file = False
            return

        level = self.levelname(level)
        h = FileHandler(path, mode=mode, encoding=encoding)

        h.setLevel(level)
        h.setFormatter(self.formatter)
        self.addHandler(h)

    def exception_handler(self, e_type, e_msg, e_traceback):
        """
        sys.excepthook = log.exception_handler

        try:
            1 / 0
        except:
            log.exception_handler(*sys.exc_info())
        """
        self.exception_message = self._formatted_exception_message(e_type, e_msg)
        self.critical(self.exception_message, exc_info=(e_type, e_msg, e_traceback))
        flush_stdout(sleep_ms=500)

        if self.exception_callback:
            self.exception_callback()

        return self.exception_message

    def _formatted_exception_message(self, e_type, e_msg):
        from multiprocessing import current_process
        from threading import current_thread

        _e_type_ = object_name(e_type) or 'Exception'
        _e_msg_  = str(e_msg).replace('"', "'")

        exception_message = '''
    
            (The result 'W:resign' was added to the game information)
            
            Error:    {e_type}: {e_msg}
            Process:  {process_name}
            Thread:   {thread_name}
            Log:      {repr_self}
            
 
        '''.format(e_type=_e_type_,
                   e_msg=_e_msg_,
                   process_name=current_process().name,
                   thread_name=current_thread().name,
                   repr_self=self.name)
        exception_message = exception_message.strip()

        _bw_ = max(len(line) for line in exception_message.split('\n')) + 8
        _bf_ = '{:*^%i}' % _bw_
        banner_upper = _bf_.format('  __Fatal Exception__  ')
        banner_lower = '*' * _bw_

        exception_message = '''
        
        {banner_upper}
            {exception_message}
        {banner_lower}
        Log Paths:
        {paths}
        

        '''.format(banner_upper=banner_upper,
                   exception_message=exception_message,
                   banner_lower=banner_lower,
                   paths=self.path_hyperlinks())
        exception_message = dedent(exception_message)

        return exception_message

    def handle(self, record: LogRecord):
        self.records.append(record)
        super().handle(record)

    def __repr__(self):
        def sort_filehandlers_last(h):
            if 'FileHandler' in h: return 2
            else:                  return 1

        ln = self.levelname()
        rh = sorted([object_name(h) for h in self.handlers],
                    key=sort_filehandlers_last)
        rh = ' | '.join(rh)

        return '{}: [{}]  handlers={{{}}}'.format(self.name, ln, rh)

    @staticmethod
    def resolve_log_path(name_or_path):
        p_path = parse_path(name_or_path, abspath=False)

        if not (p_path.directory or p_path.extension):
            return None

        if p_path.directory:
            directory = p_path.directory
        else:
            directory = standardize_path(p_path.directory, abspath=True)

        if not os.path.exists(directory):
            raise FileExistsError('log directory does not exist: \n{}'.format(directory))

        path = directory + p_path.filename + p_path.extension

        return path

    @staticmethod
    def log_formatter(log_format, date_format):
        if '%(' in log_format:
            style_format = '%'
        elif '{' in log_format:
            style_format = '{'
        else:
            style_format = '$'

        # noinspection PyTypeChecker
        return Formatter(log_format,
                         date_format,
                         style_format)


class colored_streamhandler_cls(StreamHandler):
    level_effects = {NOTSET:   None,
                     DEBUG:    'bold',
                     INFO:     'bold',
                     WARNING:  'bold',
                     ERROR:    'bold',
                     CRITICAL: 'bold'}

    level_colors  = {NOTSET:   'grey',
                     DEBUG:    'grey',
                     INFO:     'white',
                     WARNING:  'yellow',
                     ERROR:    'bright red',
                     CRITICAL: 'bright magenta'}

    def __init__(self, stream=None):
        super().__init__(stream)

    def emit(self, record):
        color  = self.level_colors.get(record.levelno)
        effect = self.level_effects.get(record.levelno)

        s = self.format(record) + self.terminator
        s = styled(s, color, effect)

        try:
            self.flush()
            self.stream.write(s)
        except RecursionError as e:
            raise e
        except Exception as e:
            print('{} Error: {}'.format(self.__class__.__name__, e))
            self.handleError(record)




import re

# for office 2010+
max_rows = 1048575           # 2**20
max_cols = 16384             # 2**14


def col_letter_offset(cs, offset):
    return col_letter(col_number(cs) + offset)


def col_letter(ci):
    """ column numbers to string """
    if isinstance(ci, str):
        return __validate_column_letter(ci)

    __validate_column_number(ci)

    cs = ''
    while ci > 0:
        ci_2 = ((ci - 1) % 26)
        cs   = chr(ci_2 + 65) + cs
        ci   = (ci - ci_2) // 26

    return cs


def col_number(cs):
    """ column letters to integer """
    if isinstance(cs, (float, int)):
        return __validate_column_number(int(cs))

    cs = __validate_column_letter(cs)

    ci = 0
    for i, ci_2 in enumerate(cs, 1):
        ci_2 = ord(ci_2) - 64
        ci  += ci_2 * 26**(len(cs) - i)

    __validate_column_number(ci)

    return ci


def __validate_column_letter(cs):
    cs = str(cs).upper()

    if not re.search('^[a-z]{1,3}$', cs, re.I):
        raise ValueError("'{}' is not a valid Excel column address "
                         "\n(valid columns should be 'A - {}')".format(cs, col_letter(max_cols)))

    return cs


def __validate_column_number(ci):
    if ci < 1:
        raise ValueError("column number ({:,}) must be >= 1".format(ci))

    if ci > max_cols:
        raise ValueError("column number ({:,}) exceeds Excel's maximum column limit".format(ci))

    return ci




import gc
import inspect
import itertools
import functools
import re
import sys

from timeit import default_timer
from time import sleep

from ..conditional import is_utf_console
from ..conditional import config

if is_utf_console:
    __vengeance_prefix__ = '    ν: '    # 'ν': chr(957), nu
else:
    __vengeance_prefix__ = '    v: '    # 'v': chr(118)

# region {ansi effect escape codes}
__effect_end__   = '\x1b[0m'
__effect_codes__ = {'bold':      '\x1b[1m',
                    'italic':    '\x1b[3m',
                    'underline': '\x1b[4m',
                    '':          ''}
__color_codes__ = {'grey':           '\x1b[29m',
                   'gray':           '\x1b[29m',
                   'white':          '\x1b[38;2;255;255;255m',
                   'red':            '\x1b[31m',
                   'green':          '\x1b[32m',
                   'yellow':         '\x1b[33m',
                   'blue':           '\x1b[34m',
                   'magenta':        '\x1b[35m',
                   'cyan':           '\x1b[36m',
                   'bronze':         '\x1b[37m',
                   'bright red':     '\x1b[91m',
                   'bright green':   '\x1b[92m',
                   'bright yellow':  '\x1b[93m',
                   # 'bright yellow':  '\x1b[33;1m',
                   'bright blue':    '\x1b[94m',
                   'bright magenta': '\x1b[95m',
                   'bright cyan':    '\x1b[96m',
                   'black':          '\x1b[97m',
                   '':               ''}
# endregion


def print_runtime(f=None,
                  color=None,
                  effect=None,
                  formatter=None,
                  end='\n'):

    def runtime_wrapper(_f_):
        @functools.wraps(_f_)
        def functools_wrapper(*args, **kwargs):
            tic  = default_timer()
            retv = _f_(*args, **kwargs)
            toc  = default_timer()
            elapsed = -(tic - toc)

            # variables for .format(**locals())
            vengeance_prefix   = __vengeance_prefix__
            formatted_function = function_name(_f_)
            formatted_elapsed  = format_seconds(elapsed)
            formatted_runtime  = '@{}: {}'.format(formatted_function, formatted_elapsed)

            if formatter is None:
                s = '{}{}'.format(vengeance_prefix, formatted_runtime)
            else:
                s = formatter.format(**locals())

            s = styled(s, color, effect)
            flush_stdout()
            print_unicode(s, end)

            return retv

        return functools_wrapper

    if (f is not None) and not callable(f):
        f, color, effect, formatter, end = None, f, color, effect, end

    if color is None:     color     = config.get('color')
    if effect is None:    effect    = config.get('effect')
    if formatter is None: formatter = config.get('formatter')

    if f:
        return runtime_wrapper(f)
    else:
        return runtime_wrapper


def print_performance(f=None, repeat=5):

    def performance_wrapper(_f_):
        @functools.wraps(_f_)
        def functools_wrapper(*args, **kwargs):
            retv  = None
            best  = None
            worst = None
            total = 0.0
            functools_repeat = functools.partial(itertools.repeat, None)

            gc_enabled = gc.isenabled()
            gc.disable()

            for _ in functools_repeat(repeat):
                tic  = default_timer()
                retv = _f_(*args, **kwargs)
                toc  = default_timer()
                elapsed = -(tic - toc)

                if best is None:
                    best  = elapsed
                    worst = elapsed
                else:
                    best  = min(best,  elapsed)
                    worst = max(worst, elapsed)

                total += elapsed

            average = total / repeat

            num_trials = surround_single_brackets(format_integer(repeat, comma_sep='_'))
            num_trials = '{} trials'.format(num_trials)
            if repeat == 1:
                num_trials = num_trials[:-1]

            s = ('@{} over {}:'
                 '\n        ★ best:     {}'
                 '\n        ☆ average:  {}'
                 '\n        ☆ worst:    {}'
                 .format(function_name(_f_),
                         num_trials,
                         format_seconds(best),
                         format_seconds(average),
                         format_seconds(worst))
                 )
            s = vengeance_message(s)

            flush_stdout()
            print_unicode(s)

            if gc_enabled:
                gc.enable()

            return retv

        return functools_wrapper

    if (f is not None) and not callable(f):
        f, repeat = None, f

    if f:
        return performance_wrapper(f)
    else:
        return performance_wrapper


def styled(message,
           color=None,
           effect=None):

    if color is None:  color  = config.get('color',  '')
    if effect is None: effect = config.get('effect', '')

    color  = (color.lower()
                   .replace('_', ' '))
    effect = (effect.lower()
                    .replace(' ', '')
                    .split('|'))

    if color not in __color_codes__:
        if not (color.startswith('\x1b') and color.endswith('m')):
            raise KeyError('styled: invalid color: {}'.format(color))

    for e in effect:
        if e not in __effect_codes__:
            if not (e.startswith('\x1b') and e.endswith('m')):
                raise KeyError('styled: invalid effect: {}'.format(e))

    if not is_utf_console:
        return message

    if not color and not effect:
        return message

    ansi_color  = __color_codes__.get(color, color)
    ansi_effect = ''.join(__effect_codes__.get(e, e) for e in effect)
    __effect_start__  = ansi_color + ansi_effect

    # only apply new style to any unstyled parts of message
    _message_ = message.replace(__effect_end__,
                                __effect_end__ + __effect_start__)

    _message_ = '{}{}{}'.format(__effect_start__,
                                _message_,
                                __effect_end__)

    return _message_


def deprecated(f=None, message='deprecated'):

    def deprecated_wrapper(_f_):
        @functools.wraps(_f_)
        def functools_wrapper(*args, **kwargs):

            _message_ = "'{}'".format(message)
            _message_ = '@{}: {}'.format(function_name(_f_), _message_)

            vengeance_warning(_message_,
                              DeprecationWarning,
                              stacklevel=4)

            return _f_(*args, **kwargs)

        return functools_wrapper

    if (f is not None) and not callable(f):
        f, message = None, f

    if callable(f):
        return deprecated_wrapper(f)
    else:
        return deprecated_wrapper


def vengeance_warning(message,
                      category=Warning,
                      stacklevel=3,
                      stackframe=None,
                      color=config.get('color'),
                      effect=config.get('effect')):
    """
    follow icecream's implementation?
        call_frame = inspect.currentframe().f_back
    """
    import traceback
    import warnings

    # region {closure functions}
    stacklevel_min = 3
    aligned_indent = ' ' * len(__vengeance_prefix__)

    def first_non_vengeance_frame():
        nonlocal stackframe

        is_found = False
        tb_stack = reversed(traceback.extract_stack())

        for i, stackframe in enumerate(tb_stack):
            if '\\vengeance\\' not in stackframe.filename:
                is_found = True
                break

        if not is_found:
            stackframe = traceback.extract_stack(inspect.currentframe(), limit=stacklevel_min)[0]

    def extract_frame():
        nonlocal stackframe

        sl = max(stacklevel, stacklevel_min)
        stackframe = traceback.extract_stack(inspect.currentframe(), limit=sl)[0]

    def vengeance_formatwarning(*_, **__):
        line_1 = ' '.join(message.split()).lstrip()

        line_1 = '<{}> {}'.format(object_name(category), line_1)
        line_1 = vengeance_message(line_1)
        line_1 = styled(line_1, color, effect)

        line_2 = 'File "{}", line {}'.format(filename, lineno)
        line_2 = styled(line_2, color, effect)
        line_2 = aligned_indent + line_2

        _message_ = '{}\n{}\n'.format(line_1, line_2)

        return _message_
    # endregion

    if stacklevel is None and stackframe is None:
        first_non_vengeance_frame()
    elif isinstance(stacklevel, int) and stackframe is None:
        extract_frame()

    if stackframe is not None:
        filename = stackframe.filename
        lineno   = stackframe.lineno
    else:
        filename = '{unknown}'
        lineno   = '{unknown}'

    flush_stdout()

    original_formatwarning = warnings.formatwarning
    warnings.formatwarning = vengeance_formatwarning
    warnings.warn(message)
    warnings.formatwarning = original_formatwarning

    flush_stdout()


def flush_stdout(sleep_ms=None):
    print(end='')
    sys.stdout.flush()

    if sleep_ms:
        sleep(sleep_ms / 1000)


def print_unicode(s, end='\n'):
    """ replace all \x1b[34m with regex? """
    try:
        print(s, end=end)
    except UnicodeError:
        print(ascii(s), end=end)


def vengeance_message(message):
    return __vengeance_prefix__ + str(message)


def surround_double_brackets(h):
    return '⟪{}⟫'.format(h)


def surround_single_brackets(h):
    return '⟨{}⟩'.format(h)


def surround_square_brackets(h):
    return '⟦{}⟧'.format(h)


def format_integer(i, comma_sep='_'):
    """ eg:
    '1_000_000' = format_integer(1000000)
    """
    _i_ = '{:,}'.format(int(i))
    return _i_.replace(',', comma_sep)


def format_seconds(s):
    s = abs(s)

    ns = s * 1e9
    us = s * 1e6
    ms = s * 1e3

    m, s = divmod(s, 60)
    h, m = divmod(m, 60)
    d, h = divmod(h, 24)

    if d >= 1:
        f_ms = '{:.0f}d {:.0f}h {:.0f}m'.format(d, h, m)
    elif h >= 1:
        f_ms = '{:.0f}h {:.0f}m {:.0f}s'.format(h, m, s)
    elif m >= 1:
        f_ms = '{:.0f}m {:.0f}s'.format(m, s)
    elif s >= 1:
        f_ms = '{:.1f} sec'.format(s)
    elif ms >= 1:
        f_ms = '{:.1f} ms'.format(ms)
    elif us >= 1:
        if is_utf_console:  f_ms = '{:.0f} μs'.format(us)
        else:               f_ms = '{:.0f} us'.format(us)
    else:
        f_ms = '{:.0f} ns'.format(ns)

    return f_ms


def function_parameters(f):
    """
    eg:
        def function(something, *, also, also_1=None):
            pass

        Arguments(args=['something', 'also', 'also_1'],
                  varargs=None,
                  varkw=None)
        = inspect.getargs(function.__code__)

        (something, *, also, also_1=None) = inspect.signature(function)

        OrderedDict([('something',  <Parameter "something">),
                     ('also',       <Parameter "also">),
                     ('also_1',     <Parameter "also_1=None">)])
        = list(inspect.signature(function).parameters.items())
    """
    # region {closure param_cls}
    class param_cls:
        __slots__ = ('name',
                     'kind',
                     'default',
                     'value')

        def __init__(self, p):
            self.name    = p.name
            self.kind    = str(p.kind)
            self.default = p.default
            self.value   = None

            if object_name(p.default) != '_empty':
                self.value = p.default

        def __repr__(self):
            return '{}={!r}  {}'.format(self.name,
                                        self.value,
                                        '{' + self.kind + '}')
    # endregion

    i_params = inspect.signature(f).parameters
    n_params = [param_cls(p) for p in i_params.values()]

    return n_params


def function_name(f):
    try:
        name = f.__qualname__
    except AttributeError:
        try:
            if isinstance(f, property): name = str(f.fget).split(' ')[1]
            else:                       name = str(f)
        except:
            name = str(f)

    if '.' in name:         # probably a class method
        return name

    try:                   modulename = f.__module__
    except AttributeError: modulename = str(f)

    modulename = modulename.split('.')[-1]

    return '{}.{}'.format(modulename, name)


def object_name(o):
    try:                   return o.__name__
    except AttributeError: pass

    try:                   return o.__class__.__name__
    except AttributeError: pass

    try:                   return type(o).__name__
    except AttributeError: return ''


# noinspection DuplicatedCode
def snake_case(s):
    """ eg:
        'some_value' = snake_case('someValue')
    """
    s: str

    camel_re = re.compile('''
        (?<=[a-z])[A-Z](?=[a-z])
    ''', re.VERBOSE)

    s = s.strip()
    if s[0].isupper():
        s = s[0].lower() + s[1:]

    for i, match in enumerate(camel_re.finditer(s)):
        c = match.span()[0] + i
        p = '_' + s[c].lower()
        s = '{}{}{}'.format(s[:c], p, s[c + 1:])

    return s.lower()








from .flux_cls import flux_cls
from .log_cls import log_cls

# following more traditional PascalCase naming convention for classes
# noinspection PyPep8Naming
from .flux_cls import flux_cls as Flux
# noinspection PyPep8Naming
from .log_cls import log_cls   as VenLog

__all__ = ['flux_cls',
           'log_cls',
           'Flux',
           'VenLog']




from typing import Dict
from ... conditional import ordereddict

'''
re-map flat dictionary keys to a nested dictionary structure

eg:
    # "flat" dict
    d = {('a₁', 'b₁', 'c₁', 'd₁'): 'v_1',
         ('a₁', 'b₂', 'c₁', 'd₂'): 'v_2',
         ('a₂', 'b₁', 'c₁', 'd₁'): 'v_4',
         ('a₂', 'b₁', 'c₁', 'd₂'): 'v_3'}

    {'a₁': {'b₁': {'c₁': {'d₁': 'v_1'}},
            'b₂': {'c₁': {'d₂': 'v_2'}}},
     'a₂': {'b₁': {'c₁': {'d₁': 'v_4',
                          'd₂': 'v_3'}}}} = tree_cls(d).traverse()
'''


class node_cls:
    def __init__(self, value=None):
        self.children = ordereddict()
        self.value    = value


class tree_cls:
    def __init__(self, flat_dict: Dict):
        if not isinstance(flat_dict, dict):
            raise TypeError('flat_dict must be a dictionary')

        self.children = ordereddict()
        for flat_keys, value in flat_dict.items():
            self.add(flat_keys, value)

    def add(self, flat_keys, value):
        node = self

        for e_key in flat_keys:
            if e_key not in node.children:
                child = node_cls(value)

                node.children[e_key] = child
                node                 = child
            else:
                node = node.children[e_key]

    def traverse(self, node=None):
        node = node or self

        d = ordereddict()
        for k, node in node.children.items():
            if not node.children:
                d[k] = node.value
            else:
                d[k] = self.traverse(node)

        return d





import os
import setuptools
from setuptools import setup

"""
Publishing (Perfect) Python Packages on PyPi
https://youtu.be/GIF3LaRqgXo

https://markdownlivepreview.com/
https://hynek.me/articles/conditional-python-dependencies/
https://betterprogramming.pub/a-python-package-developers-cheat-sheet-3efb9e9454c7

pip install vengeance[comtypes]
pip install vengeance[pypiwin32]
pip install vengeance[python-dateutil]
pip install vengeance[numpy]
pip install vengeance[comtypes,pypiwin32,python-dateutil,numpy]

todo:
    publish to conda
"""

is_windows_os = (os.name == 'nt')

__version__ = '1.1.26'
__release__ = '$release 63'
description = 'Library focusing on row-major organization of tabular data and control over the Excel application'

try:
    with open('README.md', 'r') as f:
        long_description = f.read()

    long_description_content_type = 'text/markdown'
except:
    long_description = 'https://github.com/michael-ross-ven/vengeance/blob/master/README.md'
    long_description_content_type = 'text'

if is_windows_os:
    install_requires = ['comtypes',
                        'pypiwin32']
else:
    install_requires = []

extras_require = {":python_version>='3.0'": ['python-dateutil',
                                             'numpy']}


if __name__ == '__main__':
    setup(name='vengeance',
          version=__version__,
          description=description,
          long_description=long_description,
          long_description_content_type=long_description_content_type,
          url='https://github.com/michael-ross-ven/vengeance',
          author='Michael Ross',
          author_email='',
          license='MIT',
          install_requires=install_requires,
          extras_require=extras_require,
          packages=setuptools.find_packages(),
          classifiers=[
              "Programming Language :: Python :: 3",
              "License :: OSI Approved :: MIT License"
            ]

          )
